{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE MagicHash #-}
{-# LANGUAGE ScopedTypeVariables #-}
module Engine.Eval
( EvalResult(..)
, CompiledExpression(..)
, evalWithSeq
) where

import Control.Applicative
import Control.Exception
import CoreLint
import CoreSyn
import CoreUtils (exprType)
import Data.List
import Engine.EngineHValue
import Engine.EngineM
import FastString
import GHC hiding (compileExpr, exprType)
import GHC.Prim
import HscMain (compileExpr)
import HscTypes
import Id
import InstEnv
import MkCore
import Outputable
import Prelude hiding (catch)
import TysWiredIn (unitDataConId)
import UniqSupply
import Unsafe.Coerce
import Util.GHC

data EvalResult 
  = EvalOk !EngineHValue
  | EvalException !SomeException
  | EvalString !String

data CompiledExpression = CE HValue !Type
                        | CEShow String


evalWithSeq :: CompiledExpression -> IO EvalResult
evalWithSeq (CE hval etipe) = do
  (hval `seq` do
    ehv <- liftIO $ (mkEngineHValue hval etipe)
    return $ EvalOk ehv
   ) `catch` \(e :: SomeException) -> do
    e' <- makeSafe e
    return (EvalException e')

evalWithSeq (CEShow str) = do
  (str',excn) <- firstExceptionIn str
  case excn of
    Nothing -> return (EvalString str')
    Just e  -> return (EvalException e)

firstExceptionIn :: String -> IO (String, Maybe SomeException)
firstExceptionIn str = do
  (case str of
    ((!x):xs) -> do
      (!rest, !se) <- firstExceptionIn xs
      return $! (x:rest, se)
    []  -> return ("", Nothing)) `catch` (\(e :: SomeException) -> do
                                            safeExcn <- makeSafe e
                                            return ("<!>", Just safeExcn)
                                            )
    


makeSafeStr :: String -> IO String
makeSafeStr str = do
  (case str of
    ((!x):xs) -> do
      rest <- makeSafeStr xs
      rest `seq` return (x:rest)
    []     -> return "") `catch` (\(_ :: SomeException) -> return "<nested exception thrown>")
    
makeSafe :: SomeException -> IO SomeException
makeSafe e = (do
  () <- return (foldl' seq '!' str `seq` ())
  return e) `catch` (\(_ :: SomeException) -> do
    str' <- makeSafeStr str
    return (SomeException (ErrorCall str'))
    )
  where
    str = show e
