{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE PatternGuards #-}
{-# LANGUAGE ViewPatterns #-}
{-# LANGUAGE BangPatterns #-}
{-# OPTIONS_GHC -fno-warn-overlapping-patterns #-}
module Engine.Impl
( Engine
, buildEngine
, stepEngine
, EngineStepResult(..)
, ExperimentResult(..)
) where 

import Control.Applicative
import Data.Int
import Data.Sequence (Seq)
import Engine.EngineM
import Engine.Experiment
import Engine.ExperimentCache
import Engine.ExperimentResult
import Engine.Profile
import Engine.RuntimePlan
import GHC
import MonadUtils
import Plan

data Engine a
  = Engine { currentPlan      :: !(RuntimePlan a)
           , engineState      :: !EngineState
           , engineEnv        :: !EngineEnv
           }

data EngineStepResult a
  = NoWork
  | Work { experimentRun   :: !Experiment
         , ranExperiment   :: !RanExperiment
         , nextEngine      :: !(Maybe (Engine a))
         , debugLog        :: !(Seq String)
         }


buildEngine :: PlanStrategy a => Bool -> Profile -> Maybe Strategy -> Int -> Int64 -> a -> Engine a
buildEngine useShowInstances profile useExperimentCaches maxExpressionTime maxExpressionAllocs plan 
  = Engine { currentPlan      = initRuntimePlan plan
           , engineState      = initialEngineState 
           , engineEnv        = EngineEnv { maxExpressionTime   = maxExpressionTime
                                          , maxExpressionAllocs = maxExpressionAllocs
                                          , profile             = profile
                                          , useShowInstances    = useShowInstances
                                          }
           }
  where
    initialEngineState = mkInitialEngineState useExperimentCaches

stepEngine :: PlanStrategy a => Engine a -> Ghc (EngineStepResult a)
stepEngine Engine { currentPlan = currentPlan
                  , engineState = engineState 
                  , engineEnv = engineEnv
                  } = do
  (mRes , !engineState', EngineLog debugLog) <- runEngineM engineState engineEnv (stepPlan currentPlan)
  case mRes of
    Nothing -> return NoWork
    Just ((!exp, !ranExp), !mRP) -> return 
      $! Work { experimentRun  = exp
             , ranExperiment  = ranExp
             , nextEngine       = (\currentPlan -> Engine { currentPlan = currentPlan  
                                                          , engineState = engineState'
                                                          , engineEnv   = engineEnv
                                                          }) <$> mRP
             , debugLog         = debugLog
             }
  where
    stepPlan :: PlanStrategy a => RuntimePlan a -> EngineM (Maybe ((Experiment, RanExperiment), Maybe (RuntimePlan a)))
    stepPlan !currentPlan = profileFunction "stepPlan" $ do
      -- liftIO $ putStrLn (dumpRuntimePlan currentPlan)
      (!mNewPlan, !mResult) <- updateRuntimePlan currentPlan
      case mResult of 
        Just !result -> return (Just (result, mNewPlan))
        Nothing      -> maybe (return Nothing) stepPlan mNewPlan
