{-# LANGUAGE DisambiguateRecordFields #-}
{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE NamedFieldPuns #-}
{-# LANGUAGE PatternGuards #-}
{-# LANGUAGE GADTs #-}
{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE MagicHash #-}
{-# LANGUAGE ForeignFunctionInterface #-}
{-# LANGUAGE ViewPatterns #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# OPTIONS_GHC -fno-warn-overlapping-patterns #-}
module FrontEnd.Vis 
(visAction
) where

import Bottom
import Control.Applicative
import Control.Monad
import Control.Monad.Identity
import DataCon
import Data.IORef
import Data.List
import Data.Maybe
import Data.Monoid
import Data.Sequence (Seq, (><), (<|), (|>))
import Data.Time.Clock
import Data.View
import Digraph
import Distribution.Simple.Setup
import DynFlags
import Encoding
import Engine
import Engine.EngineM
import Foreign
import Foreign.C.String
import Foreign.C.Types
import Format
import FrontEnd.Constants
import FrontEnd.Options
import FrontEnd.Stats
import GHC
import GHC.Paths
import GraphRewriting
import GraphRewriting.GL.Render
import GraphRewriting.GL.UI as UI
import GraphRewriting.Graph.Write.Unsafe as Unsafe
import GraphRewriting.Layout.Coulomb
import GraphRewriting.Layout.Gravitation
import GraphRewriting.Layout.SpringEmbedder
import GraphRewriting.Layout.Wrapper
import HscTypes
import IdDB
import Linker
import ObjLink
import Outputable
import Plan as Plan
import PrelNames (unpackCStringName)
import qualified Data.Foldable as F
import qualified Data.Sequence as Seq
import qualified GraphRewriting.Graph.Write.Unsafe as Unsafe
import qualified Util.BiMap as BiMap
import System.FilePath
import System.IO
import Trace.Hpc.Mix
import Util
import Util.BiMap (BiMap)
import Util.GHC

import Vis.Graph
import Vis.Rules
import Vis.GL()


visAction :: [String] -> [String] -> IrulanOptions -> IO ()
visAction ghcOpts targets options = do
  let args = as
        where
          as = map (mkGeneralLocated "on the commandline") (concatMap words ghcOpts)

  (argsStatic, warnsStatic) <- parseStaticFlags args 

  runGhc (Just libdir) $ do
    (dflags', _, warns) <- flip parseDynamicFlags argsStatic =<< getSessionDynFlags 
    let dflags'' = dflags' { hscTarget = HscAsm 
                           , ghcLink   = LinkBinary
                           }
    primPackages <- setSessionDynFlags dflags''
    dflags <- getSessionDynFlags
    -- TODO change from this to one that says 'irulan blew up' or something
    defaultErrorHandler dflags $ do
      defaultCleanupHandler dflags $ do
          let userSpecifiedModules = map dropExtension targets
          let userSpecifiedExplicitSources = map dropExtension (explicitIdSource options)

          liftIO $ mapM_ (print . unLoc) (warns ++ warnsStatic)

          targets <- mapM (flip guessTarget Nothing) (nub $ userSpecifiedModules ++ userSpecifiedExplicitSources)
          setTargets targets

          result <- load LoadAllTargets

          liftIO $ initDynLinker dflags
          do
            dflags <- getSessionDynFlags
            liftIO $ linkPackages dflags primPackages
            -- For some reason, unpackCStringName# doesn't always get loaded.
            -- Force it to be so here.
            withSession $ \hsc -> (liftIO $ Linker.getHValue hsc unpackCStringName >> return ())
            return ()


          if (succeeded result) 
            then do
              mg <- depanal [] False
              let mods = flattenSCCs (topSortModuleGraph False mg Nothing)

              explicitIdentSourceMods <- mapM (flip lookupModule Nothing . mkModuleName) userSpecifiedExplicitSources

              forM_ mods $ \modsum -> do
                printWarnings

                modIface <- getModIface modsum
                
                let (modName :: String) = moduleNameString . ms_mod_name $ modsum

                linkModuleAndDependencies modsum modIface
                
                when (modName `elem` userSpecifiedModules) $ do
                  liftIO $ putStrLn (modName ++ ":")

                  constants <- makeConstants (intConstants options)
                                             (charConstants options)
                                             (floatConstants options)
                                             (doubleConstants options)
                                             (integerConstants options)
                  theDB <- mkIdDBForModule (fromFlag $ useVCO options)
                                           (ms_mod modsum) 
                                           (fromFlag $ onlyTestProperties options) 
                                           (fromFlag $ sourceIdsFrom options) 
                                           explicitIdentSourceMods constants

                  visualiseFunctions options theDB 

                `gcatch` \(e :: SourceError) -> do
                           printExceptionAndWarnings e

            else do
              liftIO $ putStrLn "No!"

linkModuleAndDependencies :: ModSummary -> ModIface -> Ghc ()
linkModuleAndDependencies modsum modiface = do
    dflags <- getSessionDynFlags
    liftIO $ linkPackages dflags (dep_pkgs . mi_deps $ modiface)
    let obj_file = ml_obj_file . ms_location $ modsum
    liftIO $ loadObj obj_file
    liftIO $ (resolveObjs >> return ())
    return ()


visualiseFunctions :: IrulanOptions -> IdDB -> Ghc ()
visualiseFunctions opts iddb = do
  let useCases = fromFlag $ useCaseStatement opts
  let useCaseArgs = fromFlag $ useCaseArgument opts
  let treePlan = generateTreePlan iddb useCaseArgs useCases 

  liftIO $ UI.initialise
  
  let root = fromPlan treePlan

  let engineState = mkInitialEngineState Nothing

  reifyGhc $ \session -> do
    let env = Env { session = session
                  , engineState = engineState
                  , engineEnv = EngineEnv { maxExpressionTime   = fromFlag $ FrontEnd.Options.maxExpressionTime opts
                                          , maxExpressionAllocs = fromFlag $ FrontEnd.Options.maxExpressionAllocs opts
                                          , profile             = nullProfile
                                          , useShowInstances    = fromFlag $ FrontEnd.Options.useShowOnResult opts
                                          }
                  }
    envRef <- newIORef env
    UI.run 40 id layoutStep (wrapGraph root) (ruleTree envRef)

layoutStep n = do
	(cgf, cf, sf, rot) <- readOnly $ do
		cgf <- centralGravitation n -- this is needed in order to prevent unconnected subgraphs to drift apart infinitely
		cf <- coulombForce n        -- a repulsing force between each pair of nodes
		sf <- springForce 1.5 n     -- this tries to push nodes into the position as dictated by the edges connected to other nodes.
		rot <- angularMomentum n    -- this tries to rotate nodes such that the angular forces exercised by the edges upon the nodes are minimised.
		return (cgf, cf, sf, rot)
	-- combine all the forces specifying for each of them a function that maps the distance the force bridges to a strength.
	Unsafe.adjustNode (Position . sf (*0.9) . cgf (*0.01) . cf (\x -> min (100/(x^2+0.1)) 10) . position) n

ruleTree env = Branch "All" [Leaf "Step" (planToDo env)]

instance Render (Wrapper PlanNode) where render = render . wrappee
