module Util.UniqSet
( UniqSet
, singleton
, toList
, member
) where

import qualified UniqSet as US
import Data.Monoid
import Unique
import Outputable
import qualified Data.Foldable as F

newtype UniqSet a = US (US.UniqSet a)

instance (Eq a) => Eq (UniqSet a) where
  (US a) == (US b) = sizeA == sizeB
                  && sizeA == sizeIntersect
                  && sizeB == sizeIntersect
    where
      sizeA = US.sizeUniqSet a 
      sizeB = US.sizeUniqSet b
      intersect = US.intersectUniqSets a b
      sizeIntersect = US.sizeUniqSet intersect


instance (Outputable a) => Show (UniqSet a) where
  show (US a) = showPpr a

instance Outputable a => Outputable (UniqSet a) where
  ppr (US a) = ppr a

instance Monoid (UniqSet a) where
  mempty  = US (US.emptyUniqSet)
  (US l) `mappend` (US r) = US (US.unionUniqSets l r)
  
singleton :: Uniquable a => a -> UniqSet a
singleton = US . US.unitUniqSet

toList :: UniqSet a -> [a]
toList (US a) = US.uniqSetToList a

member :: Uniquable a => a -> UniqSet a -> Bool
member a (US s) = a `US.elementOfUniqSet` s

instance F.Foldable UniqSet where
  foldr f b (US a) = US.foldUniqSet f b a
